import json
import requests
import shutil

from functools import wraps
from pathlib import Path


class IOHandler:
    _metadata_filename = None
    client = None

    @classmethod
    def iterdir(cls, location):
        raise NotImplementedError

    @classmethod
    def to_dict(cls, location):
        raise NotImplementedError

    @classmethod
    def from_dict(cls, location, metadata, filepath=None):
        raise NotImplementedError

    @classmethod
    def touch(cls, location):
        raise NotImplementedError

    @classmethod
    def exists(cls, location):
        raise NotImplementedError

    @classmethod
    def mkdir(cls, location):
        raise NotImplementedError

    @classmethod
    def read(cls, location):
        raise NotImplementedError

    @classmethod
    def unlink(cls, location):
        raise NotImplementedError


class LocalJsonHandler(IOHandler):
    _metadata_filename = "metadata.json"

    def rewrite_location(func):
        @wraps(func)
        def wrapper(cls, location, *args, **kwargs):
            if isinstance(location, Path):
                location = str(location)
            location = location.replace("file://", "")
            return func(cls, location, *args, **kwargs)

        return wrapper

    @classmethod
    @rewrite_location
    def read(cls, location, chunk_size=8192):
        with open(Path(location), "rb") as file:
            while chunk := file.read(chunk_size):
                yield chunk

    @classmethod
    @rewrite_location
    def iterdir(cls, location):
        return [i.name for i in Path(location).iterdir()]

    @classmethod
    @rewrite_location
    def copy(cls, location, filepath):
        shutil.copy(filepath, location)
        return cls.exists(f"{location}/{Path(filepath).name}")

    @classmethod
    @rewrite_location
    def mkdir(cls, location):
        return Path(location).mkdir()

    @classmethod
    @rewrite_location
    def touch(cls, location):
        return Path(location).touch()

    @classmethod
    @rewrite_location
    def unlink(cls, location):
        return Path(location).unlink()

    @classmethod
    @rewrite_location
    def exists(cls, location):
        return Path(location).exists()

    @classmethod
    @rewrite_location
    def to_dict(cls, location, filepath=None):
        if filepath is None:
            filepath = f"{location}/{cls._metadata_filename}"
        try:
            with open(filepath.replace("file://", ""), "rt") as file:
                return json.load(file)
        except (FileNotFoundError, json.JSONDecodeError):
            return {}
        except Exception as e:
            raise e

    @classmethod
    @rewrite_location
    def from_dict(cls, location, metadata, filepath=None):
        if filepath is None:
            filepath = f"{location}/{cls._metadata_filename}"
        with open(filepath.replace("file://", ""), "w") as file:
            json.dump(metadata, file)


class HttpJsonHandler(IOHandler):
    _metadata_filename = "metadata.json"

    @classmethod
    def read(cls, location, chunk_size=8192):
        with requests.get(location, stream=True) as r:
            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=chunk_size):
                yield chunk

    @classmethod
    def iterdir(cls, location):
        return requests.get(location).json()

    @classmethod
    def copy(cls, src, dst):
        response = requests.get(src)
        with open(dst, "wb") as file:
            file.write(response.content)
        return cls.exists(dst)

    @classmethod
    def mkdir(cls, location):
        return requests.put(location)

    @classmethod
    def touch(cls, location):
        return requests.post(location)

    @classmethod
    def unlink(cls, location):
        return requests.delete(location)

    @classmethod
    def exists(cls, location):
        return requests.get(location).ok

    @classmethod
    def to_dict(cls, location, filepath=None):
        if filepath is None:
            filepath = f"{location}/{cls._metadata_filename}"
        return requests.get(f"{location}/{cls._metadata_filename}").json()

    @classmethod
    def from_dict(cls, location, metadata, filepath=None):
        if filepath is None:
            filepath = f"{location}/{cls._metadata_filename}"
        return requests.post(filepath, json=metadata)


class WebHdfsJsonHandler(IOHandler):
    _metadata_filename = "metadata.json"

    def rewrite_location(func):
        @wraps(func)
        def wrapper(cls, location, *args, **kwargs):
            prefix = cls.client.url.replace("http://", "webhdfs://")
            location = location.replace(prefix, "")
            return func(cls, location, *args, **kwargs)

        return wrapper

    @classmethod
    @rewrite_location
    def read(cls, location, chunk_size=8192):
        with cls.client.read(location, chunk_size=chunk_size) as reader:
            for chunk in reader:
                yield chunk

    @classmethod
    @rewrite_location
    def iterdir(cls, location):
        return cls.client.list(location)

    @classmethod
    @rewrite_location
    def copy(cls, location, filepath):
        cls.client.upload(location, filepath)
        return cls.exists(f"{location}/{Path(filepath).name}")

    @classmethod
    @rewrite_location
    def mkdir(cls, location):
        return cls.client.makedirs(location)

    @classmethod
    @rewrite_location
    def touch(cls, location):
        return cls.client.write(location, "")

    @classmethod
    @rewrite_location
    def unlink(cls, location):
        return cls.client.delete(location)

    @classmethod
    @rewrite_location
    def exists(cls, location):
        try:
            cls.client.status(location)
            return True
        except:
            return False

    @classmethod
    @rewrite_location
    def to_dict(cls, location, filepath=None):
        if filepath is None:
            filepath = f"{location}/{cls._metadata_filename}"
        prefix = cls.client.url.replace("http://", "webhdfs://")
        filepath = filepath.replace(prefix, "")
        try:
            return json.loads(b"".join(cls.read(filepath)).decode())
        except (FileNotFoundError, json.JSONDecodeError):
            return {}
        except Exception as e:
            raise e

    @classmethod
    @rewrite_location
    def from_dict(cls, location, metadata, filepath=None):
        if filepath is None:
            filepath = f"{location}/{cls._metadata_filename}"
        prefix = cls.client.url.replace("http://", "webhdfs://")
        filepath = filepath.replace(prefix, "")
        cls.client.write(filepath, json.dumps(metadata), overwrite=True)



import os
import json
from http.server import SimpleHTTPRequestHandler, HTTPServer


class JSONDirectoryHandler(SimpleHTTPRequestHandler):
    def list_directory(self, path):
        try:
            encoded = json.dumps(os.listdir(path)).encode("utf-8")
            self.send_response(200)
            self.send_header("Content-type", "application/json; charset=utf-8")
            self.send_header("Content-Length", str(len(encoded)))
            self.end_headers()
            self.wfile.write(encoded)
        except os.error:
            self.send_error(404, "No permission to list directory")


if __name__ == "__main__":
    port = 8000
    server_address = ("0.0.0.0", port)
    httpd = HTTPServer(server_address, JSONDirectoryHandler)
    print(f"Serving on port {port}")
    httpd.serve_forever()



from setuptools import setup, find_packages

setup(
    name="metatreedb",
    version="0.1.7",
    author="oboki",
    author_email="oboki@kakao.com",
    description="Metatree is a DBMS that uses the filesystem itself as a tree-structured database.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/oboki/metatreedb",
    packages=find_packages(),
    install_requires=["requests", "hdfs"],
    extras_require={"dev": ["pytest"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.8",
)



import asyncio
import logging
import pickle
import pytest
import shutil
import uuid

from functools import partial
from pathlib import Path

from metatree import Metatree


@pytest.fixture(scope="session")
def shared_fixture():
    basepath = f"/tmp/{uuid.uuid4().hex[:8]}"
    Path(basepath).mkdir()
    with open(Path(f"{basepath}/trained.pkl"), "wb") as file:
        pickle.dump(
            (
                "spam",
                "eggs",
            ),
            file,
        )
    metatree = Metatree(
        f"{basepath}/metatree",
        (
            "model",
            "version",
            "stage",
        ),
        locking_enabled=True,
    )
    yield metatree, basepath
    shutil.rmtree(basepath)


def test_put(shared_fixture):
    metatree, basepath = shared_fixture
    res = metatree.put(
        {"model": "model_a", "version": "v1", "stage": "training"},
        f"{basepath}/trained.pkl",
    )
    assert res == True


def test_find(shared_fixture):
    metatree, basepath = shared_fixture
    got = metatree.find({"model": "model_a", "version": "v1", "stage": "training"})
    assert got.location == f"file://{basepath}/metatree/model_a/v1/training"
    got = metatree.find({"model": "model_a", "version": "v1"})
    assert got.location == f"file://{basepath}/metatree/model_a/v1"


def test_get(shared_fixture):
    metatree, basepath = shared_fixture
    for chunk in metatree.get("model_a/v1/training/trained.pkl"):
        with open(f"{basepath}/downloaded.pkl", "ab") as f:
            f.write(chunk)
    assert Path(f"{basepath}/downloaded.pkl").exists() == True


def test_list(shared_fixture):
    metatree, _ = shared_fixture
    got = metatree.find({"model": "model_a", "version": "v1", "stage": "training"})
    assert got.list() == ["trained.pkl"]


def test_update(shared_fixture):
    metatree, _ = shared_fixture
    got = metatree.find({"model": "model_a"})
    got.update(active="v1")
    assert got.metadata.get("active") == "v1"


def test_dict_query(shared_fixture):
    metatree, basepath = shared_fixture
    got = metatree.find(
        {
            "model": "model_a",
            "version": {"metadata": "active"},
            "stage": {"value": "training"},
        }
    )
    assert got.location == f"file://{basepath}/metatree/model_a/v1/training"


def test_string_query(shared_fixture):
    metatree, basepath = shared_fixture
    got = metatree.find("model_a/<active>/training")
    assert got.location == f"file://{basepath}/metatree/model_a/v1/training"


async def async_update(mtree, **kwargs):
    loop = asyncio.get_event_loop()
    func = partial(mtree.update, **kwargs)
    await loop.run_in_executor(None, func)


async def remove_lock_file(basepath):
    await asyncio.sleep(5)
    Path(f"{basepath}/.lock").unlink()


async def _test_lock(shared_fixture):
    metatree, basepath = shared_fixture
    Path(f"{basepath}/.lock").touch()
    await asyncio.gather(
        async_update(metatree, spam="eggs"),
        remove_lock_file(basepath),
    )
    # exception_occurred = 0
    # for _ in range(3):
    #     with pytest.raises(Exception):
    #         await asyncio.gather(
    #             async_update(metatree, spam="eggs"),
    #             remove_lock_file(basepath),
    #         )
    #         exception_occurred += 1
    # assert exception_occurred == 3


def test_lock(shared_fixture, caplog):
    # logging.getLogger().setLevel(logging.DEBUG)
    # caplog.set_level(logging.WARNING)
    asyncio.run(_test_lock(shared_fixture))
    metatree, _ = shared_fixture
    assert metatree.metadata.get("spam") == "eggs"
